#include <sourcemod>
#include <sdktools>
#include <cstrike>
#include <dhooks>
#pragma semicolon 1

public Plugin myinfo =
{
    name = "CS:GO Knives",
    author = "Neuro Toxin",
    description = "A highly refined knife selection plugin",
    version = "1.0.1"
};

int g_iPlayerKnifeSelections[MAXPLAYERS+1];
Handle g_hGiveNamedItem = null;
Handle g_hGiveNamedItemPost = null;
Handle g_hKnivesMenu = null;

// This data will load via config if this plugin gets anywhere
int g_iKnifeDefinitions[8] = {500,505,506,507,508,509,512,515};
char g_sKnifeDefinitionNames[8][64] = {"Bayonet", "Flip Knife", "Gut Knife", "Karambit", "M9 Bayonet", "Huntsman Knife", "Falchion Knife", "Butterfly Knife" };
char g_sKnifeClassnames[8][64] = {"weapon_bayonet", "weapon_knife_flip", "weapon_knife_gut", "weapon_knife_karambit", "weapon_knife_m9_bayonet", "weapon_knife_tactical", "weapon_knife_falchion", "weapon_knife_butterfly"};

public void OnPluginStart()
{
	if (!HookOnGiveNamedItem())
	{
		SetFailState("Unable to hook GiveNamedItem using DHooks");
		return;
	}
	
	BuildKnivesMenu();
	RegisterConsoleCommands();
}

public void RegisterConsoleCommands()
{
	RegConsoleCmd("sm_knife", OnKnifeCommand);
}

public bool HookOnGiveNamedItem()
{
	new Handle:config = LoadGameConfigFile("knives.games");
	if(config == INVALID_HANDLE)
	{
		LogError("Unable to load game config file: knives.games");
		return false;
	}
	
	int offset = GameConfGetOffset(config, "GiveNamedItem");
	CloseHandle(config);
	if (offset == -1)
	{
		LogError("Unable to find offset 'GiveNamedItem' in game data 'knives.games'");
		return false;
	}
	
	/* POST HOOK */
	g_hGiveNamedItemPost = DHookCreate(offset, HookType_Entity, ReturnType_CBaseEntity, ThisPointer_CBaseEntity, OnGiveNamedItemPost);
	
	if (g_hGiveNamedItemPost == INVALID_HANDLE)
	{
		LogError("Unable to post hook 'int CCSPlayer::GiveNamedItem(char const*, int, CEconItemView*, bool)'");
		return false;
	}
	
	DHookAddParam(g_hGiveNamedItemPost, HookParamType_CharPtr, -1, DHookPass_ByRef);
	DHookAddParam(g_hGiveNamedItemPost, HookParamType_Int, -1, DHookPass_ByVal);
	DHookAddParam(g_hGiveNamedItemPost, HookParamType_Int, -1, DHookPass_ByRef);
	DHookAddParam(g_hGiveNamedItemPost, HookParamType_Bool, -1, DHookPass_ByVal);
	
	/* PRE HOOK */
	g_hGiveNamedItem = DHookCreate(offset, HookType_Entity, ReturnType_CBaseEntity, ThisPointer_CBaseEntity, OnGiveNamedItem);
	
	if (g_hGiveNamedItem == INVALID_HANDLE)
	{
		LogError("Unable to hook 'int CCSPlayer::GiveNamedItem(char const*, int, CEconItemView*, bool)'");
		return false;
	}
	
	DHookAddParam(g_hGiveNamedItem, HookParamType_CharPtr, -1, DHookPass_ByRef);
	DHookAddParam(g_hGiveNamedItem, HookParamType_Int, -1, DHookPass_ByVal);
	DHookAddParam(g_hGiveNamedItem, HookParamType_Int, -1, DHookPass_ByRef);
	DHookAddParam(g_hGiveNamedItem, HookParamType_Bool, -1, DHookPass_ByVal);
	return true;
}

public void OnConfigsExecuted()
{
	// Late loading here is the best place
	for (int client = 1; client < MaxClients; client++)
	{
		if (!IsClientConnected(client))
			continue;
			
		OnClientConnected(client);
		
		if (!IsClientInGame(client))
			continue;
			
		if (IsFakeClient(client))
			continue;
			
		OnClientPutInServer(client);
	}
}

public void OnClientConnected(client)
{
	g_iPlayerKnifeSelections[client] = 0;
}

public void OnClientPutInServer(client)
{
	if (IsFakeClient(client))
		return;
	
	DHookEntity(g_hGiveNamedItem, false, client);
	DHookEntity(g_hGiveNamedItemPost, true, client);
}

/*
 * This is a prehook to change the classnames of the item being spawned if required.
 * -- This stops the animation glitch of some knives
*/
public MRESReturn OnGiveNamedItem(int client, Handle hReturn, Handle hParams)
{
	// Skip if player hasnt selected a knife
	if (!g_iPlayerKnifeSelections[client])
		return MRES_Ignored;
	
	// Get classname and skip processing if it's not a weapon
	static char classname[64];
	DHookGetParamString(hParams, 1, classname, sizeof(classname));
	
	// -> not sure if leaving this this is a performance boost or not
	if (StrContains(classname, "weapon_") != 0)
		return MRES_Ignored;
	
	// Update the classname and if successful ...
	if (UpdateKnifeClassname(client, classname, sizeof(classname)))
	{
		//PrintToConsole(client, "==============-----------------> OnGiveNamedItemBlocked(classname=%s)", classname);
		
		/* The original plan was to override the classname param
		 * -> dhooks either has a bug or the game pushes the classname back to players inventory item */
		
		// Give the player the correct classnamed item
		int weapon = GivePlayerItem(client, classname);
		if (weapon > -1) EquipPlayerWeapon(client, weapon);
		DHookSetReturn(hReturn, -1);
		
		// Block this call
		// -> I would assume the entity wouldnt be created when this is blocked, however it is and the post is called :(
		return MRES_Supercede;
	}
	
	//PrintToConsole(client, "==============-----------------> OnGiveNamedItem(classname=%s)", classname);
	return MRES_Ignored;
}

public MRESReturn OnGiveNamedItemPost(int client, Handle hReturn, Handle hParams)
{
	// Skip if player hasnt selected a knife
	if (!g_iPlayerKnifeSelections[client])
		return MRES_Ignored;
	
	// Get classname and skip processing if it's not a weapon
	// -> this is done first as some named items dont contain netprop m_iItemDefinitionIndex
	static char classname[64];
	DHookGetParamString(hParams, 1, classname, sizeof(classname));
	if (StrContains(classname, "weapon_") != 0)
		return MRES_Ignored;
		
	// For some reason the orginal blocked hook appears to have its post called via dhooks
	// -> simply skip processing for the default knife classnames
	if (StrEqual(classname, "weapon_knife"))
		return MRES_Ignored;
	else if (StrEqual(classname, "weapon_knife_t"))
		return MRES_Ignored;
	
	int weapon = DHookGetReturn(hReturn);
	//PrintToConsole(client, "===============---------------------> OnGiveNamedItemPost(classname=%s, entity=%d)", classname, weapon);
	
	// Get definition index
	int defindex = GetEntProp(weapon, Prop_Send, "m_iItemDefinitionIndex");
	
	// Skip if player is not getting a knife
	bool isknife = IsDefIndexKnife(defindex);
	if (!isknife)
		return MRES_Ignored;
		
	// Skip if item definitions match
	if (g_iPlayerKnifeSelections[client] == defindex)
		return MRES_Ignored;
	
	// Set knife definition index to players selection
	SetEntProp(weapon, Prop_Send, "m_iItemDefinitionIndex", g_iPlayerKnifeSelections[client]);
	//PrintToConsole(client, "=================---------------------> forced defindex=%d", g_iPlayerKnifeSelections[client]);
	
	// Remove paintkit info
	SetEntProp(weapon, Prop_Send, "m_iItemIDLow", -1);
	SetEntProp(weapon, Prop_Send, "m_nFallbackPaintKit", 1);
	
	// I reset the seed and wear as they leak through from a players paintkit
	SetEntProp(weapon, Prop_Send, "m_nFallbackSeed", 100);
	SetEntPropFloat(weapon, Prop_Send, "m_flFallbackWear", 0.001);
	return MRES_Ignored;
}

stock bool IsDefIndexKnife(int defindex)
{
	if (defindex == 42 || defindex == 59)
		return true;
		
	int count = sizeof(g_iKnifeDefinitions);
	for (int i = 0; i < count; i++)
	{
		if (g_iKnifeDefinitions[i] == defindex)
			return true;
	}
	return false;
}

stock void BuildKnivesMenu()
{
	g_hKnivesMenu = CreateMenu(OnClientMenuAction);
	SetMenuTitle(g_hKnivesMenu, "Select Knife");
	SetMenuExitBackButton(g_hKnivesMenu, false);
	SetMenuExitButton(g_hKnivesMenu, true);
	
	char info[32];
	int count = sizeof(g_iKnifeDefinitions);
	for (int i = 0; i < count; i++)
	{
		Format(info, sizeof(info), "$knives %d", g_iKnifeDefinitions[i]);
		AddMenuItem(g_hKnivesMenu, info, g_sKnifeDefinitionNames[i]);
	}
	
	AddMenuItem(g_hKnivesMenu, "$knives 0", "Default");
}

public Action OnKnifeCommand(int client, int args)
{
	DisplayMenu(g_hKnivesMenu, client, 0);
}

public int OnClientMenuAction(Handle menu, MenuAction action, int param1, int param2)
{
	if (action == MenuAction_Select)
	{
		char buffer[64];
		GetMenuItem(menu, param2, buffer, sizeof(buffer));
	
		if (StrContains(buffer, "$knives ") != 0)
			return;
			
		int defindex = StringToInt(buffer[7]);
		char classname[64];
		if (!GetItemDefinitionDetails(defindex, buffer, sizeof(buffer), classname, sizeof(classname)))
			return;
			
		g_iPlayerKnifeSelections[param1] = defindex;
		PrintToChat(param1, "\x01[SM] You have selected knife: \x04%s", buffer);
		
		if (IsPlayerAlive(param1))
		{
			// We must remove any tasers before we equip a new knife
			bool equiptaser = false;
			int weapon = GetPlayerWeaponSlot(param1, CS_SLOT_KNIFE);
			while (weapon > -1)
			{
				if (GetEntProp(weapon, Prop_Send, "m_iItemDefinitionIndex") == 31)
					equiptaser = true;
				
				RemovePlayerItem(param1, weapon);
				AcceptEntityInput(weapon, "Kill");
				weapon = GetPlayerWeaponSlot(param1, CS_SLOT_KNIFE);
			}
			
			weapon = GivePlayerItem(param1, classname);
			if (weapon > -1) EquipPlayerWeapon(param1, weapon);
			if (equiptaser) GivePlayerItem(param1, "weapon_taser");
		}
	}
}

stock bool GetItemDefinitionDetails(int defindex, char[] name="", int namelength=0, char[] classname="", int classnamelength=0)
{
	if (defindex == 0)
	{
		Format(name, namelength, "Default");
		Format(classname, classnamelength, "weapon_knife");
		return true;
	}
	
	int count = sizeof(g_iKnifeDefinitions);
	for (int i = 0; i < count; i++)
	{
		if (g_iKnifeDefinitions[i] == defindex)
		{
			Format(name, namelength, "%s", g_sKnifeDefinitionNames[i]);
			Format(classname, classnamelength, "%s", g_sKnifeClassnames[i]);
			return true;
		}
	}
	return false;
}

stock bool UpdateKnifeClassname(int client, char[] classname, length)
{
	// Add these two manually so they are not displayed in the menu
	if (StrEqual(classname, "weapon_knife"))
		return GetItemDefinitionDetails(g_iPlayerKnifeSelections[client], _, _, classname, length);
	else if (StrEqual(classname, "weapon_knife_t"))
		return GetItemDefinitionDetails(g_iPlayerKnifeSelections[client], _, _, classname, length);
	else
	{
		int count = sizeof(g_sKnifeClassnames);
		for (int i = 0; i < count; i++)
		{
			if (!StrEqual(g_sKnifeClassnames[i], classname))
				continue;
				
			if (g_iPlayerKnifeSelections[client] != g_iKnifeDefinitions[i])
				return GetItemDefinitionDetails(g_iPlayerKnifeSelections[client], _, _, classname, length);
		}
	}
	return false;
}